@@ -9,10 +9,71 @@ from rohub import rohub
 
 class Resource(object):
     """
-    Class Representation of rohub's Resource.
+    Class Representation of Rohub's resource.
+
+    .. note::
+
+        | **editable attributes:**
+        | - resource_type
+        | - title
+        | - folder
+        | - description
+
+        | **read-only attributes:**
+        | - ros
+        | - identifier
+        | - name
+        | - filename
+        | - size
+        | - download_url
+        | - created
+        | - creator
+        | - modificator
+        | - modified
+        | - created_on
+        | - created_by
+        | - modified_on
+        | - modified_by
+        | - original_created_on
+        | - original_created_by
+        | - original_creator_name
+        | - authors_credits
+        | - contributors_credits
+        | - shared
+        | - doi
+        | - read_only
+        | - api_link
     """
+
     def __init__(self, ro_id=None, source=None, resource_type=None, input_url=None, file_path=None,
                  title=None, folder=None, description=None, identifier=None, post_request=True):
+        """
+        Constructor for the Resource.
+
+        .. seealso::
+            :func:`~rohub.list_valid_resource_types`
+
+        :param ro_id: research object's identifier, optional
+        :type ro_id: str
+        :param source: resource's source - either internal or external, optional
+        :type source: str
+        :param resource_type: type of resource, optional
+        :type resource_type: str
+        :param input_url: resource's url, optional
+        :type input_url: str
+        :param file_path: resource's file path, optional
+        :type file_path: str
+        :param title: resource's title, optional
+        :type title: str
+        :param folder: folder's identifier, optional
+        :type folder: str
+        :param description: resource's description
+        :type description: str
+        :param identifier: resource's identifier, optional, used when object already exists and has to be loaded
+        :type identifier: str
+        :param post_request: if True, the object will be created, otherwise loaded - default is True
+        :type post_request: bool
+        """
         if self._is_valid():
             # Main.
             self.res_response_content = {}
@@ -20,14 +81,14 @@ class Resource(object):
 
             if post_request:
                 # Required attributes
-                self.ro_id = ro_id   # REGULAR
-                self.resource_type = resource_type   # VALUE VALIDATION
-                self.input_url = input_url   # REGULAR
-                self.file_path = file_path   # REGULAR
+                self.ro_id = ro_id  # REGULAR
+                self.resource_type = self._validate_resource_type(res_type=str(resource_type))
+                self.input_url = input_url  # REGULAR
+                self.file_path = file_path  # REGULAR
                 # Optional attributes
-                self.title = title   # TYPE VALIDATION
-                self.folder = folder   # TYPE VALIDATION
-                self.description = description   # TYPE VALIDATION
+                self.title = title  # TYPE VALIDATION
+                self.folder = folder  # TYPE VALIDATION
+                self.description = description  # TYPE VALIDATION
                 # Crating new Research Object.
                 self._post_resource()
             else:
@@ -46,7 +107,7 @@ class Resource(object):
 
             # Updating optional attributes with values from the response.
             self.title = self.res_response_content.get("title")
-            self.folder = self.res_response_content.get("folder")
+            self.folder = utils.map_folder_id_to_path(self.res_response_content.get("folder"))
             self.description = self.res_response_content.get("description")
 
             # ReadOnly attributes; will be updated after request post.
@@ -96,17 +157,6 @@ class Resource(object):
     ###############################################################################
     #              Properties.                                                    #
     ###############################################################################
-    @property
-    def resource_type(self):
-        return self.__resource_type
-
-    @resource_type.setter
-    def resource_type(self, value):
-        if value:
-            self.__resource_type = self._validate_resource_type(res_type=str(value))
-        else:
-            self.__resource_type = None
-
     @property
     def title(self):
         return self.__title
@@ -331,10 +381,15 @@ class Resource(object):
     @staticmethod
     def _is_valid():
         """
-        Semi-Private function that checks if current token is still valid.
+        Semi-Private function that checks if current token is still valid, and if not it attempts
+        to refresh it.
         :return: boolean -> True if valid, False otherwise.
         """
-        return utils.is_valid(token_type="access")
+        if utils.is_valid(token_type="access"):
+            valid_token = True
+        else:
+            valid_token = utils.refresh_access_token()
+        return valid_token
 
     def _post_resource(self):
         """
@@ -384,8 +439,10 @@ class Resource(object):
 
     def show_metadata(self):
         """
-        Function that shows selected (most relevant) information regarding Resource.
-        :return: dict -> dictionary containing most relevant information regarding Resource.
+        Function that shows selected (most relevant) information regarding resource.
+
+        :returns: resource's most relevant metadata
+        :rtype: dict
         """
         basic_metadata = {
             "identifier": self.identifier,
@@ -406,17 +463,43 @@ class Resource(object):
 
     def show_full_metadata(self):
         """
-        Function that shows all meta data associated with the Resource.
-        :return: JSON -> response containing all meta-data for the Resource.
+        Function that shows all metadata associated with the resource.
+
+        :returns: response content from the API
+        :rtype: dict
         """
         return self.res_response_content
 
-    def update_metadata(self):
+    def update_metadata(self, ro_id=None):
         """
-        Function for updating Resource metadata in the service.
-        :return: JSON -> response.
+        Function that updates resource's metadata.
+
+        .. note::
+            After executing update_metadata the resource will be updated in the service with accordance
+            to the changes that were made to the python object in your local scope.
+
+        .. note::
+            to move back a resource to the "Home" directory set folder = "/"
+
+        :param ro_id: research object's identifier, optional
+        :type ro_id: str
+        :returns: response content from the API
+        :rtype: dict
         """
+        if self.folder:
+            if not ro_id:
+                if len(self.ros) > 1:
+                    msg = "This resource is associated with more than one research object, please explicitly provide " \
+                          "research object id (ro_id) to identify the changes in the folders structure " \
+                          " to a particular research object"
+                    raise SystemExit(msg)
+                else:
+                    ro_id = self.ros[0]
+        # validating attributes before making a call to the API
+        self.resource_type = self._validate_resource_type(res_type=str(self.resource_type))
+        # making a call
         self.update_metadata_response_content = rohub.resource_update_metadata(identifier=self.identifier,
+                                                                               ro_id=ro_id,
                                                                                res_type=self.resource_type,
                                                                                title=self.title,
                                                                                folder=self.folder,
@@ -425,11 +508,27 @@ class Resource(object):
 
     def update_content(self, input_url=None, file_path=None):
         """
-        Function for updating Resource content in the service. NOTE: change input_url attribute in case
-        of external resource, or file_path attribute for the internal.
-        :param input_url: str -> input url.
-        :param file_path: str -> path to the file.
-        :return: JSON -> response.
+        Function that updates resource's content.
+
+        .. note::
+            After executing update_content the resource will be updated in the service with accordance
+            to the changes that were made to the python object in your local scope.
+            In order to update content for internal resource please provide resource through file_path argument.
+            In case of external resource an input_url argument should be provided.
+
+        .. note::
+            input_url is required for updating external resource, and file_path
+            is in case of internal one. At least one of them has to be provided!
+
+        .. warning::
+            The resource content will be overwritten!
+
+        :param input_url: url to the external resource
+        :type input_url: str
+        :param file_path: path to the internal resource
+        :type file_path: str
+        :returns: response content from the API
+        :rtype: dict
         """
         self.update_content_response_content = rohub.resource_update_content(identifier=self.identifier,
                                                                              input_url=input_url,
@@ -438,31 +537,83 @@ class Resource(object):
 
     def delete(self):
         """
-        Function that deletes Resource.
-        :return: JSON -> response.
+        Function that deletes resource.
+
+        .. warning::
+            The resource will be deleted from the service and will no longer appear in the API.
+            This doesn't mean that object created in your local scope will be removed!
+
+        :returns: response content from the API
+        :rtype: dict
         """
         self.delete_response_content = rohub.resource_delete(identifier=self.identifier)
         return self.delete_response_content
 
-    def assign_doi(self):
+    def assign_doi(self, external_doi=None):
         """
-        Function that adds doi to the current Resource.
-        :return: str -> doi.
+        Function that assigns doi to the resource.
+
+        .. note::
+            If external_doi value was provided it will be used as doi, otherwise the system
+            will automatically generate and assign doi!
+
+        :param external_doi: value for the external, existing doi, optional
+        :type external_doi: str
+        :returns: doi
+        :rtype: str
         """
-        self.assign_doi_response_content = rohub.resource_assign_doi(identifier=self.identifier)
+        self.assign_doi_response_content = rohub.resource_assign_doi(identifier=self.identifier,
+                                                                     external_doi=external_doi)
         return self.assign_doi_response_content
 
     def download(self, resource_filename, path=None):
         """
-        Function that acquires a current Resource to the local file system.
-        :param resource_filename: str -> full filename with extension associated to the format!
-        :param path: str -> path where file should be downloaded. If nothing is specified current working
-        directory will be used.
-        :return: full path to the file.
+        Function that acquires the resource to the local file storage.
+
+        :param resource_filename: resource's full filename (with extension)
+        :type resource_filename: str
+        :param path: path where file should be downloaded, optional - current working dir is the default!
+        :type path: str
+        :returns: full path to the acquired resource
+        :rtype: str
         """
         return rohub.resource_download(identifier=self.identifier, resource_filename=resource_filename,
                                        path=path)
 
+    def set_license(self, license_id):
+        """
+        Function that sets license information to the resource.
+
+        .. seealso::
+            :func:`~list_available_licenses`
+            :func:`~list_custom_licenses`
+            :func:`~add_custom_license`
+
+        :param license_id: license's identifier
+        :type license_id: str
+        :returns: response content from the API
+        :rtype: dict
+        """
+        return rohub.resource_set_license(res_id=self.identifier, license_id=license_id)
+
+    def list_license(self):
+        """
+        Function that lists license associated with the resource.
+
+        :returns: response containing license details
+        :rtype: dict
+        """
+        return rohub.resource_list_license(identifier=self.identifier)
+
+    def delete_license(self):
+        """
+        Function that deletes association between license and resource.
+
+        :returns: response content from the API
+        :rtype: dict
+        """
+        return rohub.resource_delete_license(identifier=self.identifier)
+
     ###############################################################################
     #              Required Attributes methods.                                   #
     ###############################################################################
