@@ -1,6 +1,7 @@
 import requests
 from datetime import datetime, timedelta
 import time
+import json
 
 from rohub import settings
 from rohub import rohub
@@ -22,7 +23,7 @@ def valid_to(exp_time, token_type):
 def is_valid(token_type):
     """
     Function that checks if given token is still valid.
-    :param token_type: str -> token type.
+    :param token_type: str -> token type, either access or refresh
     :return: boolean -> True if valid, False otherwise.
     """
     if token_type.lower() == "access":
@@ -34,7 +35,7 @@ def is_valid(token_type):
             else:
                 return True
         else:
-            print("Missing information regarding token expiration time. Consider logging again!")
+            print("Missing information regarding token expiration time. Please login again!")
     elif token_type.lower() == "refresh":
         if settings.REFRESH_TOKEN_VALID_TO:
             now = datetime.now()
@@ -44,7 +45,7 @@ def is_valid(token_type):
             else:
                 return True
         else:
-            print("Missing information regarding token expiration time. Consider logging again!")
+            print("Missing information regarding token expiration time. Please login again!")
     else:
         print("Token type not recognized! Supported values are access and refresh.")
 
@@ -196,6 +197,36 @@ def post_request(url, data):
     return r
 
 
+def post_request_no_data(url):
+    """
+    Function that performs post request (without data) with error handling.
+    :param url: str -> url.
+    :return: Response object -> response.
+    """
+    r = None   # initialize request as empty
+    headers = {'Authorization': f"{settings.TOKEN_TYPE.capitalize()} {settings.ACCESS_TOKEN}"}
+    try:
+        r = requests.post(url=url, headers=headers, timeout=settings.TIMEOUT)
+        r.raise_for_status()
+    except requests.exceptions.HTTPError as e:
+        print("Http error occurred.")
+        raise SystemExit(e.response.text)
+    except requests.exceptions.ConnectionError:
+        try:
+            print("Connection error occurred. Trying again...")
+            r = requests.post(url=url, headers=headers, timeout=settings.TIMEOUT)
+            r.raise_for_status()
+        except requests.exceptions.ConnectionError as e:
+            print("Connection error occurred for a second time. Aborting...")
+            raise SystemExit(e.response.text)
+    except requests.exceptions.Timeout as e:
+        print("Timeout. Could not connect to the server.")
+        raise SystemExit(e.response.text)
+    except requests.exceptions.RequestException as e:
+        raise SystemExit(e)
+    return r
+
+
 def post_request_with_file(url, file):
     """
     Function that performs post request for uploading file with error handling.
@@ -292,6 +323,38 @@ def patch_request(url, data):
     return r
 
 
+def patch_request_json_payload(url, data):
+    """
+    Function that perform patch request with json as input payload.
+    :param url: str -> url.
+    :param data: dict -> input data.
+    :return: Response object -> response.
+    """
+    r = None   # initialize request as empty
+    headers = {'Authorization': f"{settings.TOKEN_TYPE.capitalize()} {settings.ACCESS_TOKEN}",
+               'content-type': "application/json; charset=UTF-8"}
+    try:
+        r = requests.patch(url=url, headers=headers, data=json.dumps(data), timeout=settings.TIMEOUT)
+        r.raise_for_status()
+    except requests.exceptions.HTTPError as e:
+        print("Http error occurred.")
+        raise SystemExit(e.response.text)
+    except requests.exceptions.ConnectionError:
+        try:
+            print("Connection error occurred. Trying again...")
+            r = requests.patch(url=url, headers=headers, data=json.dumps(data), timeout=settings.TIMEOUT)
+            r.raise_for_status()
+        except requests.exceptions.ConnectionError as e:
+            print("Connection error occurred for a second time. Aborting...")
+            raise SystemExit(e.response.text)
+    except requests.exceptions.Timeout as e:
+        print("Timeout. Could not connect to the server.")
+        raise SystemExit(e.response.text)
+    except requests.exceptions.RequestException as e:
+        raise SystemExit(e)
+    return r
+
+
 def patch_request_with_file(url, data, file):
     """
     Function that perform patch request.
@@ -470,8 +533,8 @@ def check_for_status(job_url):
         job_r = get_request(url=job_url, use_token=True)
         job_content = job_r.json()
         job_status = job_content['status']
-        if job_status == "SUCCESS":
-            return True
+        if job_status != "PENDING":
+            return job_status
     return False
 
 
@@ -597,21 +660,45 @@ def does_agent_exist(agent_name, allow_org):
     :return: agent's internal id if exists, None otherwise.
     """
     if isinstance(agent_name, dict):
-        agent_name = agent_name.get("org_id") or agent_name.get("user_id")
-        if not agent_name:
-            msg = "Username not found in the input data!" \
-                  " Username should be provided as either string, or dictionary that contains " \
-                  " user_id or org_id key with username value!"
+        # first possibility is to identify entity based on orcid/ror id.
+        agent_candidate_1 = agent_name.get("orcid_id") or agent_name.get("ror_identifier")
+        # second possibility is to identify entity based on email
+        agent_candidate_2 = agent_name.get("email")
+        if not agent_candidate_1 and not agent_candidate_2:
+            msg = "Missing data! When providing agent data, please make sure" \
+                  " that you are providing either orcid_id/ror_identifier fields for " \
+                  "user/organization respectively or email! At least one of those has" \
+                  " to be provided!"
             raise SystemExit(msg)
-    user_id = does_user_exist(username=agent_name)
-    if user_id:
-        return user_id
-    else:
+        # checking if any of candidates matches users/organization database record
+        if agent_candidate_1:
+            user_id = does_user_exist(username=generate_orcid_url(agent_candidate_1))
+            if user_id:
+                return user_id
+        if agent_candidate_2:
+            user_id = does_user_exist(username=agent_candidate_2)
+            if user_id:
+                return user_id
         if allow_org:
-            org_id = does_organization_exist(organization=agent_name)
-            if org_id:
+            if agent_candidate_1:
+                org_id = does_organization_exist(organization=generate_ror_url(agent_candidate_1))
+                if org_id:
+                    return org_id
+            if agent_candidate_2:
+                org_id = does_organization_exist(organization=agent_candidate_2)
                 return org_id
-        return False
+            return False
+        else:
+            return False
+    else:
+        user_id = does_user_exist(username=agent_name)
+        if user_id:
+            return user_id
+        else:
+            if allow_org:
+                org_id = does_organization_exist(organization=agent_name)
+                return org_id
+            return False
 
 
 def list_validated_agents(agents, allow_org):
@@ -655,57 +742,188 @@ def create_agent(data, user_only):
             ERROR: incorrect input_data!
             Input data has to be provided as list of dictionaries in case of non-existing agents! Below you can find
             examples of input_data for user and organization.
-            EITHER user_id or org_id has to be provided! Rest is optional. Display name will be set to user_id/
-            org_id in case it was not provided.
+            agent_type, display_name are required along with at least one of the following: orcid_id/ror_identifier, email.
             USER:
-            {"user_id":"example_username", "display_name": "example_display_name", "email":"example_email", 
-            "orcid_id":"example_orcid_id", "affiliation": "example_affiliation"}
+            {"agent_type": "user", "display_name": "example_display_name", "email": "example_email", 
+            "orcid_id": "example_orcid_id", "affiliation": "example_affiliation"}
             ORGANIZATION:
-            {"org_id":"example_organization_name", "display_name": "example_display_name", 
-            "email":"example_email", "organization_url": "example_url", "ror_identifier": "example_ror"}
+            {"agent_type": "organization", "display_name": "example_display_name", "email": "example_email", 
+            "organization_url": "example_url", "ror_identifier": "example_ror"}
             """
         else:
             msg = """
             ERROR: incorrect input_data!
             Input data has to be provided as list of dictionaries in case of non-existing agents! Below you can find
             example of input_data for user.
-            EITHER user_id has to be provided! Rest is optional. Display name will be set to user_id
-            in case it was not provided.
+            display_name is required along with at least one of the following: orcid_id, email.
             USER:
-            {"user_id":"example_username", "display_name": "example_display_name", "email":"example_email", 
+            {"display_name": "example_display_name", "email":"example_email", 
             "orcid_id":"example_orcid_id", "affiliation": "example_affiliation"}
             """
         raise SystemExit(msg)
     if user_only:
         # check and create external user only
-        if data.get("user_id"):
-            # create new external user
-            user_id = data.get("user_id")
-            display_name = data.get("display_name", user_id)
-            email = data.get("email")
-            orcid_id = data.get("orcid_id")
-            affiliation = data.get("affiliation")
-            return rohub.external_user_add(user_id=user_id, display_name=display_name, email=email,
-                                           orcid_id=orcid_id, affiliation=affiliation)
+        display_name = data.get("display_name")
+        if not display_name:
+            msg = "Exiting... required argument display_name is missing!"
+            raise SystemExit(msg)
+        email = data.get("email")
+        orcid_id = data.get("orcid_id")
+        affiliation = data.get("affiliation")
+        return rohub.external_user_add(display_name=display_name, email=email,
+                                       orcid_id=orcid_id, affiliation=affiliation)
     else:
+        agent_type = data.get("agent_type")
+        if not agent_type:
+            msg = """
+            required argument - agent_type is missing! Please check examples of proper input data
+            below:
+            USER:
+            {"agent_type": "user", "display_name": "example_display_name", "email":"example_email", 
+            "orcid_id":"example_orcid_id", "affiliation": "example_affiliation"}
+            ORGANIZATION:
+            {"agent_type": "organization", "display_name": "example_display_name", "email":"example_email", 
+            "organization_url": "example_url", "ror_identifier": "example_ror"}
+            """
+            raise SystemExit(msg)
         # check and create external user or organization.
-        if data.get("user_id"):
+        if agent_type == "user":
             # create new external user
-            user_id = data.get("user_id")
-            display_name = data.get("display_name", user_id)
+            display_name = data.get("display_name")
+            if not display_name:
+                msg = "Exiting... required argument display_name is missing!"
+                raise SystemExit(msg)
             email = data.get("email")
             orcid_id = data.get("orcid_id")
             affiliation = data.get("affiliation")
-            return rohub.external_user_add(user_id=user_id, display_name=display_name, email=email,
+            return rohub.external_user_add(display_name=display_name, email=email,
                                            orcid_id=orcid_id, affiliation=affiliation)
-        elif data.get("org_id"):
+        elif agent_type == "organization":
             # create new organization
-            org_id = data.get("org_id")
-            display_name = data.get("display_name", org_id)
+            display_name = data.get("display_name")
             email = data.get("email")
             org_url = data.get("organization_url")
             ror_id = data.get("ror_identifier")
-            return rohub.organization_add(organization_id=org_id, display_name=display_name,
-                                          email=email, organization_url=org_url, ror_identifier=ror_id)
+            return rohub.organization_add(display_name=display_name, email=email,
+                                          organization_url=org_url,
+                                          ror_identifier=ror_id)
+        else:
+            msg = "agent_type not recognized, has to be either user or organization!"
+            raise SystemExit(msg)
 
 
+def generate_orcid_url(orcid_id):
+    """
+    Helper function for generating orcid url.
+
+    :param orcid_id: str -> orcid id
+    :return: orcid url
+    """
+    return f"https://orcid.org/{orcid_id}"
+
+
+def generate_ror_url(ror_id):
+    """
+    Helper function for generating ror url.
+
+    :param ror_id: str -> ror id
+    :return: ror url
+    """
+    return f"https://ror.org/{ror_id}"
+
+
+def refresh_access_token():
+    """
+    Helper function that uses refresh token to acquire new access token in case the previous one expired.
+
+    :return: True if token was successfully refresh, False otherwise
+    """
+    if settings.REFRESH_TOKEN:
+        if is_valid(token_type="refresh"):
+            url = settings.KEYCLOAK_URL
+            data = {'client_id': settings.KEYCLOAK_CLIENT_ID,
+                    'client_secret': settings.KEYCLOAK_CLIENT_SECRET,
+                    'refresh_token': settings.REFRESH_TOKEN,
+                    'grant_type': 'refresh_token'}
+            try:
+                r = requests.post(url, data=data, timeout=settings.TIMEOUT)
+                r.raise_for_status()
+                r_json = r.json()
+                settings.ACCESS_TOKEN = r_json.get('access_token')
+                settings.ACCESS_TOKEN_VALID_TO = valid_to(exp_time=r_json.get('expires_in'),
+                                                          token_type="access token")
+                settings.REFRESH_TOKEN = r_json.get('refresh_token')
+                settings.REFRESH_TOKEN_VALID_TO = valid_to(exp_time=r_json.get('refresh_expires_in'),
+                                                           token_type="refresh token")
+                settings.TOKEN_TYPE = r_json.get('token_type')
+                settings.SESSION_STATE = r_json.get('session_state')
+                if settings.ACCESS_TOKEN:
+                    return True
+            except requests.exceptions.HTTPError as e:
+                print("Http error occurred.")
+                print(e.response.text)
+                raise e
+            except requests.exceptions.ConnectionError:
+                try:
+                    print("Connection error occurred. Trying again...")
+                    requests.post(url, data=data, timeout=settings.TIMEOUT)
+                except requests.exceptions.ConnectionError as e:
+                    print("Connection error occurred for a second time. Aborting...")
+                    raise SystemExit(e.response.text)
+                except requests.exceptions.RequestException as e:
+                    raise SystemExit(e)
+            except requests.exceptions.Timeout as e:
+                print("Timeout. Could not connect to the server.")
+                raise SystemExit(e.response.text)
+            except requests.exceptions.RequestException as e:
+                raise SystemExit(e)
+        else:
+            msg = "Your refresh token expired! Please use login function to " \
+                  "authenticate yourself again."
+            raise SystemExit(msg)
+    else:
+        if settings.KEYCLOAK_CLIENT_ID and settings.KEYCLOAK_CLIENT_SECRET:
+            rohub.login(client_id=settings.KEYCLOAK_CLIENT_ID,
+                        client_secret=settings.KEYCLOAK_CLIENT_SECRET)
+            return True
+        else:
+            print("Service account couldn't automatically re-login due to the missing"
+                  " credentials. Please use login function explicitly to regain access.")
+
+
+def map_path_to_folder_id(folder_path, ro_identifier):
+    """
+    Helper function that maps folder's path to its identifier using relation to the specific research object.
+
+    :param folder_path: str -> folder's path
+    :param ro_identifier: str -> research object's identifier
+    :return: str -> folder's identifier
+    """
+    if folder_path:
+        df = rohub.ros_list_folders(identifier=ro_identifier)
+        try:
+            folder_id = df.loc[:, "identifier"][df["path"] == folder_path].values[0]
+            return folder_id
+        except KeyError:
+            print("Aborted...")
+            print("It looks like there are no existing folders associated with this research object!")
+            return
+        except IndexError:
+            print("Aborted...")
+            print("Sorry, folder with provided path doesn't exist. Please make sure you are"
+                  " sending a correct path!")
+            return
+
+
+def map_folder_id_to_path(folder_id):
+    """
+    Helper function that maps folder's identifier to its full path.
+
+    :param folder_id: str -> folder's id
+    :return: str -> folder's full path
+    """
+    if folder_id:
+        url = settings.API_URL + f"folders/{folder_id}"
+        r = get_request(url=url, use_token=False)
+        content = r.json()
+        return content.get("path")
